# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import pytest
from mock import patch, MagicMock
import itertools as it
import numpy as np
from allensdk.api.queries.reference_space_api import ReferenceSpaceApi as RSA


@pytest.fixture
def ref_space():
    rsa = RSA()
    
    return rsa


@pytest.fixture
def mock_nrrd():
    mocked_nrrd = MagicMock()
    mocked_nrrd.read = MagicMock(return_value=('mock_annotation_data',
                                               'mock_annotation_image'))
    return mocked_nrrd


def CCF_VERSIONS():
    return [RSA.CCF_2015,
            RSA.CCF_2016,
            RSA.CCF_2017]


def DATA_PATHS(): 
    return [RSA.AVERAGE_TEMPLATE,
            RSA.ARA_NISSL,
            RSA.MOUSE_2011,
            RSA.DEVMOUSE_2012,
            RSA.CCF_2015,
            RSA.CCF_2016, 
            RSA.CCF_2017]


def RESOLUTIONS():
    return [RSA.VOXEL_RESOLUTION_10_MICRONS,
            RSA.VOXEL_RESOLUTION_25_MICRONS,
            RSA.VOXEL_RESOLUTION_50_MICRONS,
            RSA.VOXEL_RESOLUTION_100_MICRONS]

MOCK_ANNOTATION_DATA = 'mock_annotation_data'
MOCK_ANNOTATION_IMAGE = 'mock_annotation_image'



def test_download_mouse_atlas_volume(ref_space):

    with patch.object(ref_space, 'retrieve_file_over_http') as mock_retrieve:
        with pytest.raises(RuntimeError):
            ref_space.download_mouse_atlas_volume('P56', 'Mouse_gridAnnotation', 'P56/gridAnnotation.mhd')    

    mock_retrieve.assert_called_once_with(
        'http://download.alleninstitute.org/informatics-archive/'\
        'current-release/mouse_annotation/'\
        'P56_Mouse_gridAnnotation.zip',
        'P56/gridAnnotation.mhd',
        zipped=True)


@pytest.mark.parametrize("data_path,resolution",
                         it.product(DATA_PATHS(),
                                    RESOLUTIONS()))
def test_download_volumetric_data(ref_space,
                                  data_path,
                                  resolution):
    cache_filename = "annotation_%d.nrrd" % (resolution)

    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        ref_space.download_volumetric_data(data_path,
                                           cache_filename,
                                           resolution)

    mock_retrieve.assert_called_once_with(
        "http://download.alleninstitute.org/informatics-archive/"
        "current-release/mouse_ccf/%s/annotation_%d.nrrd" % 
        (data_path,
        resolution),
        cache_filename)


@pytest.mark.parametrize("resolution",
                         RESOLUTIONS())
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch('os.makedirs')
def test_download_structure_mask(os_makedirs,
                                 nrrd_read,
                                 ref_space, 
                                 resolution):

    structure_id = 12

    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        a, b = ref_space.download_structure_mask(structure_id,
                                                 None, resolution,
                                                 '/path/to/foo.nrrd',
                                                 reader=nrrd_read)

    assert a
    assert b

    expected = 'http://download.alleninstitute.org/informatics-archive/'\
               'current-release/mouse_ccf/{0}/structure_masks/'\
               'structure_masks_{1}/structure_{2}.nrrd'.format(RSA.CCF_VERSION_DEFAULT, 
                                                               resolution, 
                                                               structure_id)
    mock_retrieve.assert_called_once_with(expected, '/path/to/foo.nrrd')
    os_makedirs.assert_any_call('/path/to')


@patch('allensdk.core.obj_utilities.read_obj', return_value=('mock_obj'))
@patch('os.makedirs')
def test_download_structure_mesh(os_makedirs,
                                 read_obj,
                                 ref_space):

    structure_id = 12

    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        a = ref_space.download_structure_mesh(structure_id,
                                              None, '/path/to/foo.obj',
                                              reader=read_obj)
    
    assert a == 'mock_obj'
    
    expected = 'http://download.alleninstitute.org/informatics-archive/'\
               'current-release/mouse_ccf/{0}/structure_meshes/'\
               '{1}.obj'.format(RSA.CCF_VERSION_DEFAULT, structure_id)

    mock_retrieve.assert_called_once_with(expected, '/path/to/foo.obj')
    os_makedirs.assert_any_call('/path/to')


@pytest.mark.parametrize("ccf_version,resolution",
                         it.product(CCF_VERSIONS(),
                                    RESOLUTIONS()))
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch('os.makedirs')
def test_download_annotation_volume(os_makedirs,
                                    nrrd_read,
                                    ref_space,
                                    ccf_version,
                                    resolution):
    cache_file = '/path/to/annotation_%d.nrrd' % (resolution)

    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        ref_space.download_annotation_volume(
            ccf_version,
            resolution,
            cache_file,
            reader=nrrd_read)

    nrrd_read.assert_called_once_with(cache_file)

    mock_retrieve.assert_called_once_with(
        "http://download.alleninstitute.org/informatics-archive/"
        "current-release/mouse_ccf/%s/annotation_%d.nrrd" % 
        (ccf_version, resolution),
        "/path/to/annotation_%d.nrrd" % (resolution))

    os_makedirs.assert_any_call('/path/to')


@pytest.mark.parametrize("resolution",
                         RESOLUTIONS())
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch('os.makedirs')
def test_download_annotation_volume_default(os_makedirs,
                                            nrrd_read,
                                            ref_space,
                                            resolution):
    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        a, b = ref_space.download_annotation_volume(
            None,
            resolution,
            '/path/to/annotation_%d.nrrd' % (resolution),
            reader=nrrd_read)

    assert a
    assert b

    print(mock_retrieve.call_args_list)

    mock_retrieve.assert_called_once_with(
        "http://download.alleninstitute.org/informatics-archive/"
        "current-release/mouse_ccf/%s/annotation_%d.nrrd" % 
        (RSA.CCF_VERSION_DEFAULT, resolution),
        "/path/to/annotation_%d.nrrd" % (resolution))

    os_makedirs.assert_any_call('/path/to')


@pytest.mark.parametrize("resolution",
                         RESOLUTIONS())
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch('os.makedirs')
def test_download_template_volume(os_makedirs,
                                  nrrd_read,
                                  ref_space,
                                  resolution):
    with patch.object(ref_space, "retrieve_file_over_http") as mock_retrieve:
        ref_space.download_template_volume(
            resolution,
            '/path/to/average_template_%d.nrrd' % (resolution),
            reader=nrrd_read)

    mock_retrieve.assert_called_once_with(
        "http://download.alleninstitute.org/informatics-archive/"
        "current-release/mouse_ccf/average_template/average_template_%d.nrrd" % 
        (resolution),
        "/path/to/average_template_%d.nrrd" % (resolution))

    os_makedirs.assert_any_call('/path/to')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import numpy as np


def read_obj(path):
    with open(path, 'r') as obj_file:
        lines = obj_file.read().split('\n')
        output = parse_obj(lines)
    return output


def parse_obj(lines):
    '''Parse a wavefront obj file into a triplet of vertices, normals, and faces. 
    This parser is specific to obj files generated from our annotation volumes

    Parameters
    ----------
    lines : list of str
        Lines of input obj file

    Returns
    -------
    vertices : np.ndarray
        Dimensions are (nSamples, nCoordinates=3). Locations in the reference space
        of vertices
    vertex_normals : np.ndarray
        Dimensions are (nSample, nElements=3). Vectors normal to vertices.
    face_vertices : np.ndarray
        Dimensions are (sample, nVertices=3). References are given in indices 
        (0-indexed here, but 1-indexed in the file) of vertices that make up each face.
    face_normals : np.ndarray
        Dimensions are (sample, nNormals=3). References are given in indices 
        (0-indexed here, but 1-indexed in the file) of vertex normals that make up each face.

    Notes
    -----
    This parser is specialized to the obj files that the Allen Institute for Brain Science 
    generates from our own structure annotations.
    '''

    vertices = []
    vertex_normals = []
    face_vertices = []
    face_normals = []

    for line in lines:
        
        if line[:2] == 'v ':
            vertices.append( line.split()[1:] )

        elif line[:3] == 'vn ':
            vertex_normals.append( line.split()[1:] )

        elif line[:2] == 'f ':
            line = line.replace('//', ' ').split()[1:]

            face_vertices.append( line[::2] )
            face_normals.append( line[1::2] )
            
    vertices = np.array(vertices).astype(float)
    vertex_normals = np.array(vertex_normals).astype(float)
    face_vertices = np.array(face_vertices).astype(int) - 1
    face_normals = np.array(face_normals).astype(int) - 1

    return vertices, vertex_normals, face_vertices, face_normals



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os

import pytest
import mock
import numpy as np
import nrrd
import pandas as pd

from allensdk.core.reference_space_cache import ReferenceSpaceCache
from allensdk.core.structure_tree import StructureTree


@pytest.fixture()
def rsp_version():
    return 'annotation/look_a_version'


@pytest.fixture()
def resolution():
    return 25


@pytest.fixture(scope='function')
def old_nodes():

    return [{'id': 0, 'structure_id_path': '/0/', 
             'color_hex_triplet': '000000', 'acronym': 'rt', 
             'name': 'root', 'parent_structure_id': 12}]


@pytest.fixture(scope='function')
def new_nodes():

    return [{'id': 0, 'structure_id_path': '/0/', 
             'color_hex_triplet': '000000', 'acronym': 'rt', 
             'name': 'root', 'structure_sets':[{'id': 1}, {'id': 4}, {'id': 167587189}] }]


@pytest.fixture(scope='function')
def rsp(fn_temp_dir, rsp_version, resolution):

    manifest_path = os.path.join(fn_temp_dir, 'manifest.json')
    return ReferenceSpaceCache(reference_space_key=rsp_version,
                               resolution=resolution, 
                               manifest=manifest_path)



def test_init(rsp, fn_temp_dir):

    manifest_path = os.path.join(fn_temp_dir, 'manifest.json')
    assert( os.path.exists(manifest_path) )


def test_get_annotation_volume(rsp, fn_temp_dir, rsp_version, resolution):

    eye = np.eye(100)
    path = os.path.join(fn_temp_dir, rsp_version, 'annotation_{0}.nrrd'.format(resolution))

    rsp.api.retrieve_file_over_http = lambda a, b: nrrd.write(b, eye)
    obtained, _ = rsp.get_annotation_volume()

    rsp.api.retrieve_file_over_http = mock.MagicMock()
    rsp.get_annotation_volume()

    rsp.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) ) 
    assert( os.path.exists(path) )


def test_get_template_volume(rsp, fn_temp_dir, resolution):

    eye = np.eye(100)
    path = os.path.join(fn_temp_dir, 'average_template_{0}.nrrd'.format(resolution))

    rsp.api.retrieve_file_over_http = lambda a, b: nrrd.write(b, eye)
    obtained, _ = rsp.get_template_volume()

    rsp.api.retrieve_file_over_http = mock.MagicMock()
    rsp.get_template_volume()

    rsp.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_structure_tree(rsp, fn_temp_dir, new_nodes):

    path = os.path.join(fn_temp_dir, 'structures.json')

    with mock.patch('allensdk.api.queries.ontologies_api.'
                    'OntologiesApi.model_query', 
                    return_value=new_nodes) as p:

        obtained = rsp.get_structure_tree()

        rsp.get_structure_tree()
        p.assert_called_once()

    assert(obtained.node_ids()[0] == 0)
    
    cm_obt = obtained.get_colormap()
    assert(len(cm_obt[0]) == 3)

    assert( os.path.exists(path) )


def test_get_reference_space(rsp, new_nodes):

    tree = StructureTree(StructureTree.clean_structures(new_nodes))
    rsp.get_structure_tree = lambda *a, **k: tree

    annot = np.arange(125).reshape((5, 5, 5))
    rsp.get_annotation_volume = lambda *a, **k: (annot, 'foo')

    rsp_obt = rsp.get_reference_space()

    assert( np.allclose(rsp_obt.resolution, [25, 25, 25]) )
    assert( np.allclose( rsp_obt.annotation, annot ) ) 


def test_get_structure_mask(rsp, fn_temp_dir, rsp_version):
  
    sid = 12

    eye = np.eye(100)
    path = os.path.join(fn_temp_dir, rsp_version, 'structure_masks', 
                        'resolution_25', 'structure_{0}.nrrd'.format(sid))

    rsp.api.retrieve_file_over_http = lambda a, b: nrrd.write(b, eye)
    obtained, _ = rsp.get_structure_mask(sid)

    rsp.api.retrieve_file_over_http = mock.MagicMock()
    rsp.get_structure_mask(sid)

    rsp.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) ) 
    assert( os.path.exists(path) )


def test_get_structure_mesh(rsp, fn_temp_dir, rsp_version):
  
    sid = 12

    path = os.path.join(fn_temp_dir, rsp_version, 'structure_meshes','structure_{0}.obj'.format(sid))

    def write_obj(path):
        with open(path, 'w') as fil:
          fil.write('vn 1 2 4')

    expected = [1, 2, 4]

    rsp.api.retrieve_file_over_http = lambda a, b: write_obj(b)
    obtained = rsp.get_structure_mesh(sid)

    rsp.api.retrieve_file_over_http = mock.MagicMock()
    rsp.get_structure_mesh(sid)

    rsp.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained[1], expected) ) 
    assert( os.path.exists(path) )


@pytest.mark.parametrize('inp,fails', [(1, False), 
                                        (pd.Series([2]), False), 
                                        ('qwerty', True)])
def test_validate_structure_id(inp, fails):

    if fails:
        with pytest.raises(ValueError) as exc:
            ReferenceSpaceCache.validate_structure_id(inp)
    else:
        out = ReferenceSpaceCache.validate_structure_id(inp)
        assert( out == int(inp) )


@pytest.mark.parametrize('inp,fails', [([1, 2, 3], False), 
                                        ([pd.Series([2]), pd.Series([3])], False), 
                                        (['qwerty', 1], True)])
def test_validate_structure_ids(inp, fails):

    if fails:
        with pytest.raises(ValueError) as exc:
            ReferenceSpaceCache.validate_structure_ids(inp)
    else:
        out = ReferenceSpaceCache.validate_structure_ids(inp)
        assert( out == list(map(int, inp)) )



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging
import os
import argparse
import matplotlib.pyplot as plt
import warnings
import h5py
import numpy as np
from functools import partial
from scipy.ndimage.filters import median_filter

from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet

GAUSSIAN_MAD_STD_SCALE = 1.4826


def movingmode_fast(x, kernelsize, y):
    """Compute the windowed mode of an array.  A running mode is initialized
    with a histogram of values over the initial kernelsize/2 values.  The mode
    is then updated as the kernel moves by adding and subtracting values from
    the histogram.

    Parameters
    ----------
    x : np.ndarray
        Array to be analyzed
    kernelsize : int
        Size of the moving window
    y : np.ndarray
        Output array to store the results
    """

    # offset so that the trace is non-negative
    minval = min(x.min(), 0)
    if minval < 0:
        x = x - minval

    maxval = x.max()

    # compute a histogram of a half kernel
    halfsize = int(kernelsize / 2)
    histo = np.bincount(np.rint(x[:halfsize]).astype(
        np.uint32), minlength=int(maxval + 2))

    # find the mode of the first half kernel
    mode = np.argmax(histo)

    # here initial mode is available
    for m in range(0, halfsize):
        q = int(round(x[halfsize + m]))

        histo[q] += 1

        if histo[q] > histo[mode]:
            mode = q

        y[m] = mode

    for m in range(halfsize, x.shape[0] - halfsize):
        p = int(round(x[m - halfsize]))
        histo[p] -= 1

        # need to find possibly new mode value
        if p == mode:
            mode = np.argmax(histo)

        q = int(round(x[m + halfsize]))

        histo[q] += 1

        if histo[q] > histo[mode]:
            mode = q

        y[m] = mode

    for m in range(x.shape[0] - halfsize, x.shape[0]):
        p = int(round(x[m - halfsize]))
        histo[p] -= 1

        # need to find possibly new mode value
        if p == mode:
            mode = np.argmax(histo)

        y[m] = mode

    # undo the offset
    if minval < 0:
        y += minval

    return 0


def movingaverage(x, kernelsize, y):
    """Compute the windowed average of an array.

    Parameters
    ----------
    x : np.ndarray
        Array to be analyzed
    kernelsize : int
        Size of the moving window
    y : np.ndarray
        Output array to store the results
    """

    halfsize = int(kernelsize / 2)
    sumkernel = np.sum(x[0:halfsize])
    for m in range(0, halfsize):
        sumkernel = sumkernel + x[m + halfsize]
        y[m] = sumkernel / (halfsize + m)

    sumkernel = np.sum(x[0:kernelsize])
    for m in range(halfsize, x.shape[0] - halfsize):
        sumkernel = sumkernel - x[m - halfsize] + x[m + halfsize]
        y[m] = sumkernel / kernelsize

    for m in range(x.shape[0] - halfsize, x.shape[0]):
        sumkernel = sumkernel - x[m - halfsize]
        y[m] = sumkernel / (halfsize - 1 + (x.shape[0] - m))

    return 0


def plot_onetrace(dff, fc):
    """Debug plotting function"""
    qs = np.rint(np.linspace(0, len(dff), 5)).astype(int)

    dff_max = dff.max()
    dff_min = dff.min()
    fc_max = fc.max()
    fc_min = fc.min()

    for qi in range(len(qs) - 1):
        r = qs[qi], qs[qi + 1]

        frames = np.arange(r[0], r[1])
        ax = plt.subplot(len(qs), 1, qi + 1)
        ax.plot(frames, dff[r[0]:r[1]], 'g')
        ax.set_ylim(dff_min, dff_max)
        ax.set_xlim(r[0], r[1])
        ax.set_xlabel('frames', fontsize=18)
        ax.set_ylabel('DF/F', fontsize=18, color='g')

        ax = ax.twinx()
        ax.plot(frames, fc[r[0]:r[1]], 'b')
        ax.set_ylim(fc_min, fc_max)
        ax.set_xlim(r[0], r[1])
        ax.set_ylabel('FC', fontsize=18, color='b')

    return 0


def compute_dff_windowed_mode(traces,
                              mode_kernelsize=5400,
                              mean_kernelsize=3000):
    """Compute dF/F of a set of traces using a low-pass windowed-mode operator.

    The operation is basically:

        T_mm = windowed_mean(windowed_mode(T))

        T_dff = (T - T_mm) / T_mm

    Parameters
    ----------
    traces : np.ndarray
       2D array of traces to be analyzed.
    mode_kernelsize : int
        Window size to use for windowed_mode.
    mean_kernelsize : int
        Window size to use for windowed_mean.

    Returns
    -------
    dff : np.ndarray
        2D array of dF/F traces.
    """
    if mode_kernelsize >= traces.shape[1]:
        mode_kernelsize = traces.shape[1] // 2
        logging.warning("Changing mode_kernelsize to " + str(mode_kernelsize))

    if mean_kernelsize >= traces.shape[1]:
        mean_kernelsize = traces.shape[1] // 4
        logging.warning("Changing mean_kernelsize to " + str(mean_kernelsize))

    if mode_kernelsize == 0 or mean_kernelsize == 0:
        raise ValueError("Kernel length is 0!")

    logging.debug("trace matrix shape: %d %d" %
                  (traces.shape[0], traces.shape[1]))

    modeline = np.zeros(traces.shape[1])
    modelineLP = np.zeros(traces.shape[1])
    dff = np.zeros((traces.shape[0], traces.shape[1]))

    logging.debug("computing df/f")

    for n in range(0, traces.shape[0]):
        if np.any(np.isnan(traces[n])):
            logging.warning(
                "trace for roi %d contains NaNs, setting to NaN", n)
            dff[n, :] = np.nan
            continue

        movingmode_fast(traces[n, :], mode_kernelsize, modeline[:])
        movingaverage(modeline[:], mean_kernelsize, modelineLP[:])
        dff[n, :] = (traces[n, :] - modelineLP[:]) / modelineLP[:]

        logging.debug("finished trace %d/%d" % (n + 1, traces.shape[0]))

    return dff


def compute_dff_windowed_median(traces,
                                median_kernel_long=5401,
                                median_kernel_short=101,
                                noise_stds=None,
                                n_small_baseline_frames=None,
                                **kwargs):
    """Compute dF/F of a set of traces with median filter detrending.

    The operation is basically:

        T_long = windowed_median(T) # long timescale kernel

        T_dff1 = (T - T_long) / elementwise_max(T_long, noise_std(T))

        T_short = windowed_median(T_dff1) # short timescale kernel

        T_dff = T_dff1 - elementwise_min(T_short, 2.5*noise_std(T_dff1))

    Parameters
    ----------
    traces : np.ndarray
       2D array of traces to be analyzed.
    median_kernel_long : int
        Window size to use for long timescale median detrending.
    median_kernel_short : int
        Window size to use for short timescale median detrending.
    noise_stds : list
        List that will contain noise_std(T_dff1) for each trace. The
        value for each trace will be appended to the list if provided.
    n_small_baseline_frames : list
        List that will contain the number of frames for each trace where
        the long-timescale median window is less than noise_std(T). The
        value for each trace will be appended to the list if provided.
    kwargs:
        Additional keyword arguments are passed to :func:`noise_std` .

    Returns
    -------
    dff : np.ndarray
        2D array of dF/F traces.
    """
    _check_kernel(median_kernel_long, traces.shape[1])
    _check_kernel(median_kernel_short, traces.shape[1])

    dff_traces = np.copy(traces)

    for dff in dff_traces:
        sigma_f = noise_std(dff, **kwargs)

        # long timescale median filter for baseline subtraction
        tf = median_filter(dff, median_kernel_long, mode='constant')
        dff -= tf
        dff /= np.maximum(tf, sigma_f)

        if n_small_baseline_frames is not None:
            n_small_baseline_frames.append(np.sum(tf <= sigma_f))

        sigma_dff = noise_std(dff, **kwargs)
        if noise_stds is not None:
            noise_stds.append(sigma_dff)

        # short timescale detrending
        tf = median_filter(dff, median_kernel_short, mode='constant')
        tf = np.minimum(tf, 2.5*sigma_dff)
        dff -= tf

    return dff_traces


def _check_kernel(kernel_size, data_size):
    if kernel_size % 2 == 0 or kernel_size <= 0 or kernel_size >= data_size:
        raise ValueError("Invalid kernel length {} for data length {}. Kernel "
                         "length must be positive and odd, and less than data "
                         "length.".format(kernel_size, data_size))


def noise_std(x, noise_kernel_length=31, positive_peak_scale=1.5,
              outlier_std_scale=2.5):
    """Robust estimate of the standard deviation of the trace noise."""
    _check_kernel(noise_kernel_length, len(x))
    if any(np.isnan(x)):
        return np.NaN
    x = x - median_filter(x, noise_kernel_length, mode='constant')
    # first pass removing big pos peak outliers
    x = x[x < positive_peak_scale*np.abs(x.min())]
    rstd = robust_std(x)
    # second pass removing remaining pos and neg peak outliers
    x = x[abs(x) < outlier_std_scale*rstd]
    return robust_std(x)


def robust_std(x):
    """Robust estimate of standard deviation.

    Estimate of the standard deviation using the median absolute
    deviation of x.
    """
    median_absolute_deviation = np.median(np.abs(x - np.median(x)))
    return GAUSSIAN_MAD_STD_SCALE*median_absolute_deviation


def calculate_dff(traces, dff_computation_cb=None, save_plot_dir=None):
    """Apply dF/F computation to a set of traces.

    The default computation method is :func:`compute_dff_windowed_median`
    using default window parameters.

    Parameters
    ----------
    traces : np.ndarray
        2D array of traces to be analyzed.
    dff_computation_cb : function
        Function that takes traces as an argument and returns an array
        of the same shape that is the calculated dF/F.
    save_plot_dir : str
        Directory to save dF/F plots to. By default no plots are saved.

    Returns
    -------
    dff : np.ndarray
        2D array of dF/F traces.
    """
    if dff_computation_cb is None:
        dff_computation_cb = compute_dff_windowed_median

    dff = dff_computation_cb(traces)

    if save_plot_dir is not None:
        if not os.path.exists(save_plot_dir):
            os.makedirs(save_plot_dir)

        for n in range(0, traces.shape[0]):
            if np.any(np.isnan(traces[n])):
                continue

            fig = plt.figure(figsize=(150, 40))
            plot_onetrace(dff[n, :], traces[n, :])

            plt.title('ROI ' + str(n) + ' ', fontsize=18)
            fig.savefig(os.path.join(save_plot_dir, 'dff_%d.png' %
                                     n), orientation='landscape')
            plt.close(fig)

    return dff


def compute_dff(traces,
                save_plot_dir=None,
                mode_kernelsize=5400,
                mean_kernelsize=3000):
    """Compute dF/F of a set of traces using a low-pass windowed-mode operator.

    This method is deprecated. Use :func:`calculate_dff` with
    dff_computation_cb = :func:`compute_dff_windowed_mode` .

    The operation is basically:

        T_mm = windowed_mean(windowed_mode(T))

        T_dff = (T - T_mm) / T_mm

    Parameters
    ----------
    traces: np.ndarray
       2D array of traces to be analyzed

    Returns
    -------
    np.ndarray with the same shape as the input array.
    """
    warnings.warn(
        FutureWarning("The default computation for dff has been changed. Use"
                      " `calculate_dff` to compute dff now."))
    computation_cb = partial(compute_dff_windowed_mode,
                             mode_kernelsize=mode_kernelsize,
                             mean_kernelsize=mean_kernelsize)
    return calculate_dff(traces, dff_computation_cb=computation_cb,
                         save_plot_dir=save_plot_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_h5")
    parser.add_argument("output_h5")
    parser.add_argument("--plot_dir")
    parser.add_argument("--log_level", default=logging.INFO)

    args = parser.parse_args()

    logging.getLogger().setLevel(args.log_level)

    # read from "data"
    if args.input_h5.endswith("nwb"):
        timestamps, traces = BrainObservatoryNwbDataSet(
            args.input_h5).get_corrected_fluorescence_traces()
    else:
        input_h5 = h5py.File(args.input_h5, "r")
        traces = input_h5["data"].value
        input_h5.close()

    dff = calculate_dff(traces, save_plot_dir=args.plot_dir)

    # write to "data"
    output_h5 = h5py.File(args.output_h5, "w")
    output_h5["data"] = dff
    output_h5.close()


if __name__ == "__main__":
    main()



import subprocess as sp
import os

import pytest


EXAMPLE_DIR = os.path.join(
    os.path.dirname(__file__),
    '..',
    '..',
    'doc_template',
    'examples_root',
    'examples'
)
EXAMPLES = [filename for filename in os.listdir(EXAMPLE_DIR) if filename.split('.')[-1] == 'py']


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true', reason="partial testing")
@pytest.mark.parametrize('script_name', EXAMPLES)
def test_inline_examples(script_name, tmpdir_factory):

    data_dir = tmpdir_factory.mktemp('inline_examples_data')
    sp.check_call(['python', os.path.join(EXAMPLE_DIR, script_name)], cwd=str(data_dir))


# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import numpy as np

import pytest
from mock import MagicMock, mock_open, patch

from allensdk.core.obj_utilities import read_obj, parse_obj


@pytest.fixture
def wavefront_obj():
    return '''

v 8578 5484.96 5227.57
v 8509.2 5487.54 5237.07
v 8564.38 5522.13 5220.41
v 8631.93 5497.82 5228.33
v 8517.88 5542.95 5234.53
v 8615.26 5563.22 5224.48

# i'm a comment!

vn -0.0247061 -0.352726 -0.935401
vn -0.235489 -0.190095 -0.953105
vn -0.0880336 -0.0323767 -0.995591
vn 0.122706 -0.209891 -0.969994
vn -0.343738 0.217978 -0.913416
vn 0.0753706 0.16324 -0.983703

I should be a comment, but am not

f 1//1 2//2 3//3 
f 4//4 1//1 3//3 
f 3//3 2//2 5//5 
f 6//6 3//3 5//5 

    '''


def test_read_obj(wavefront_obj):

    path = 'path!'

    # need to patch the version in allensdk.api.cache because of import x from y syntax above
    with patch( 'allensdk.core.obj_utilities.open', mock_open(read_data=wavefront_obj), create=True ) as p:
        obt = read_obj(path)
        p.assert_called_with(path, 'r')
        assert( obt is not None )        


def test_parse_obj(wavefront_obj):

    lines = wavefront_obj.split('\n')
    vertices, vertex_normals, face_vertices, face_normals = parse_obj(lines)
    
    assert(np.allclose( face_vertices, face_normals ))
    assert(np.allclose( face_vertices[2, :], [2, 1, 4] ))
    assert(np.allclose( vertices[1, :], [8509.2, 5487.54, 5237.07] ))
    assert(np.allclose( vertex_normals[2, :], [-0.0880336, -0.0323767, -0.995591] ))



import pytest, os
from mock import patch, mock_open, MagicMock
from allensdk.api.queries.cell_types_api import CellTypesApi

@pytest.fixture
def mock_cells_api():
    return [
        { 
            'cell_reporter_status': "fish",
            'csl__x': 1,
            'csl__y': 2,
            'csl__z': 3,
            'donor__species': 'taco',
            'specimen__id': 10,
            'specimen__name': 'joe',
            'structure__layer': 'fifteen',
            'structure_parent__id': 2,
            'structure_parent__acronym': 'ASAP',
            'line_name': 'bezier',
            'tag__dendrite_type': 'spikey',
            'tag__apical': 'stumpy',
            'nr__reconstruction_type': 'fancy',
            'donor__disease_state': 'influenza',
            'donor__id': 1,
            'specimen__hemisphere': 'hi',
            'csl__normalized_depth': 1
        },{

            'cell_reporter_status': "nofish",
            'csl__x': 1,
            'csl__y': 2,
            'csl__z': 3,
            'donor__species': 'taco',
            'specimen__id': 10,
            'specimen__name': 'joe',
            'structure__layer': 'fifteen',
            'structure_parent__id': 2,
            'structure_parent__acronym': 'ASAP',
            'line_name': 'bezier',
            'tag__dendrite_type': 'spikey',
            'tag__apical': 'stumpy',
            'nr__reconstruction_type': None,
            'donor__disease_state': None,
            'donor__id': 1,
            'specimen__hemisphere': 'hi',
            'csl__normalized_depth': 1
        }
    ]
    
    
@pytest.fixture
def mock_cells():
    return [
        { 
            'specimen_tags': [],
            'neuron_reconstructions': [],
            'data_sets': [],
            'donor': {
                'transgenic_lines': [],
                'organism': { 'name': CellTypesApi.MOUSE },
                'conditions': [ { 'name': 'disease categories - influenza' } ]
                }
            },
        { 
            'specimen_tags': [],
            'neuron_reconstructions': [],
            'data_sets': [ {} ],
            'donor': {
                'transgenic_lines': [ { 'transgenic_line_type_name': 'driver', 'name': 'fish' } ],
                'organism': { 'name': 'fish' }
                }
            },
        { 
            'specimen_tags': [],
            'neuron_reconstructions': [ {} ],
            'data_sets': [],
            'cell_reporter': { 'name': 'bob' },
            'donor': {
                'transgenic_lines': [],
                'organism': { 'name': CellTypesApi.HUMAN },
                'conditions': [ { 'name': 'disease categories - cheese' } ]
                }
            },
        ]

@pytest.fixture
def cell_types_api():
    endpoint = None
    
    if 'TEST_API_ENDPOINT' in os.environ:
        endpoint = os.environ['TEST_API_ENDPOINT']
        return CellTypesApi(endpoint)
    else:
        return None

@pytest.mark.skipif(cell_types_api() is None, reason='No TEST_API_ENDPOINT set.')
def test_list_cells_unmocked(cell_types_api):
    from allensdk.config import enable_console_log
    enable_console_log()

    # this test will always require the latest warehouse
    cells = cell_types_api.list_cells()


def test_list_cells_mocked(mock_cells):
    with patch.object(CellTypesApi, "model_query", return_value=mock_cells):
        ctapi = CellTypesApi()

        cells = ctapi.list_cells()
        assert len(cells) == 3

        flu_cells = [ cell for cell in cells if cell['disease_categories'] == [('influenza')] ]
        assert len(flu_cells) == 1
    
        cells = ctapi.list_cells(require_reconstruction=True)
        assert len(cells) == 1
    
        cells = ctapi.list_cells(require_morphology=True)
        assert len(cells) == 1

        cells = ctapi.list_cells(reporter_status=['bob'])
        assert len(cells) == 1

        cells = ctapi.list_cells(species=['HOMO SAPIENS'])
        assert len(cells) == 1

        cells = ctapi.list_cells(species=['mus musculus'])
        assert len(cells) == 1

def test_list_cells_api_mocked(mock_cells_api):
    with patch.object(CellTypesApi, "model_query", return_value=mock_cells_api):
        ctapi = CellTypesApi()

        cells = ctapi.list_cells_api()
        assert len(cells) == 2

        fcells = ctapi.filter_cells_api(cells, require_reconstruction=True)
        assert len(fcells) == 1

        fcells = ctapi.filter_cells_api(cells, require_morphology=True)
        assert len(fcells) == 1

        fcells = ctapi.filter_cells_api(cells, species=['taco'])
        assert len(fcells) == 2

        fcells = ctapi.filter_cells_api(cells, reporter_status=['fish'])
        assert len(fcells) == 1
        

    



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import allensdk.core.json_utilities as ju
import pytest
from mock import patch, MagicMock, call
import numpy as np


@pytest.fixture
def dict_obj():
    int_array, y = np.meshgrid(np.arange(2), np.arange(2))
    float_array = y.astype(float)/4.2
    bool_array =  int_array > 0

    object = {"string": "test string",
              "float_array": float_array,
              "int_array": int_array,
              "bool_array": bool_array,
              "list": ["this", "is", 1, "list"]}
    return object


def test_write_integer_array(dict_obj):
    s_in = ju.write_string({ "int_array": dict_obj["int_array"] })
    s_out = """{
  "int_array": [
    [
      0,
      1
    ],
    [
      0,
      1
    ]
  ]
}"""

    assert s_in == s_out

def test_write_float_array(dict_obj):
    s_in = ju.write_string({ "float_array": dict_obj["float_array"] })
    s_out ="""{
  "float_array": [
    [
      0.0,
      0.0
    ],
    [
      0.23809523809523808,
      0.23809523809523808
    ]
  ]
}"""

    assert s_in == s_out

def test_write_string(dict_obj):
    s_in = ju.write_string({ "string": dict_obj["string"] })
    s_out = """{
  "string": "test string"
}"""

    assert s_in == s_out

def test_write_bool_array(dict_obj):
    s_in = ju.write_string({ "bool_array": dict_obj["bool_array"] })
    s_out = """{
  "bool_array": [
    [
      false,
      true
    ],
    [
      false,
      true
    ]
  ]
}"""
    assert s_in == s_out

def test_write_list(dict_obj):
    s_in = ju.write_string({ "list": dict_obj["list"] })
    s_out = """{
  "list": [
    "this",
    "is",
    1,
    "list"
  ]
}"""
    assert s_in == s_out



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2018. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from allensdk.core import sitk_utilities
from allensdk.api.cache import Cache, cacheable

from .reference_space_api import ReferenceSpaceApi
from .grid_data_api import GridDataApi
from .rma_pager import pageable



class MouseAtlasApi(ReferenceSpaceApi, GridDataApi):
    ''' Downloads Mouse Brain Atlas grid data, reference volumes, and metadata.
    '''

    MOUSE_ATLAS_PRODUCTS = (1,)
    DEVMOUSE_ATLAS_PRODUCTS = (3,)
    MOUSE_ORGANISM = (2,)
    HUMAN_ORGANISM = (1,)

    @cacheable()
    @pageable(num_rows=2000, total_rows='all')
    def get_section_data_sets(self, gene_ids=None, product_ids=None, **kwargs):
        ''' Download a list of section data sets (experiments) from the Mouse Brain
        Atlas project.

        Parameters
        ----------
        gene_ids : list of int, optional
            Filter results based on the genes whose expression was characterized 
            in each experiment. Default is all.
        product_ids : list of int, optional
            Filter results to a subset of products. Default is the Mouse Brain Atlas.

        Returns
        -------
        list of dict : 
            Each element is a section data set record, with one or more gene 
            records nested in a list. 

        '''
        
        if product_ids is None:
            product_ids = list(self.MOUSE_ATLAS_PRODUCTS)
        criteria = 'products[id$in{}]'.format(','.join(map(str, product_ids)))

        if gene_ids is not None:
            criteria += ',genes[id$in{}]'.format(','.join(map(str, gene_ids)))

        order = kwargs.pop('order', ['\'id\''])

        return self.model_query(model='SectionDataSet', 
                                criteria=criteria,
                                include='genes',
                                order=order,
                                **kwargs)

    @cacheable()
    @pageable(num_rows=2000, total_rows='all')
    def get_genes(self, organism_ids=None, chromosome_ids=None, **kwargs):
        ''' Download a list of genes

        Parameters
        ----------
        organism_ids : list of int, optional
            Filter genes to those appearing in these organisms. Defaults to mouse (2).
        chromosome_ids : list of int, optional
            Filter genes to those appearing on these chromosomes. Defaults to all.

        Returns
        -------
        list of dict:
            Each element is a gene record, with a nested chromosome record (also a dict).

        '''

        if organism_ids is None:
            organism_ids = list(self.MOUSE_ORGANISM)
        criteria = '[organism_id$in{}]'.format(','.join(map(str, organism_ids)))

        if chromosome_ids is not None:
            criteria += ',[chromosome_id$in{}]'.format(','.join(map(str, chromosome_ids)))
        
        order = kwargs.pop('order', ['\'id\''])

        return self.model_query(model='Gene', 
                                criteria=criteria,
                                include='chromosome',
                                order=order,
                                **kwargs)

    @cacheable(strategy='create', 
               reader = sitk_utilities.read_ndarray_with_sitk, 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_expression_density(self, path, experiment_id):
        self.download_gene_expression_grid_data(
            experiment_id, GridDataApi.DENSITY, path)


    @cacheable(strategy='create', 
               reader = sitk_utilities.read_ndarray_with_sitk, 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_expression_energy(self, path, experiment_id):
        self.download_gene_expression_grid_data(
            experiment_id, GridDataApi.ENERGY, path)


    @cacheable(strategy='create', 
               reader = sitk_utilities.read_ndarray_with_sitk, 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_expression_intensity(self, path, experiment_id):
        self.download_gene_expression_grid_data(
            experiment_id, GridDataApi.INTENSITY, path)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging
import os
from ..biophys_sim.neuron.hoc_utils import HocUtils
from allensdk.core.nwb_data_set import NwbDataSet
from fractions import gcd
from skimage.measure import block_reduce
import scipy.interpolate
import numpy as np
from pkg_resources import resource_filename #@UnresolvedImport

PERISOMATIC_TYPE = "Biophysical - perisomatic"
ALL_ACTIVE_TYPE = "Biophysical - all active"


def create_utils(description, model_type=None):
    ''' Factory method to create a Utils subclass.

    Parameters
    ----------
    description : Config instance
        used to initialize Utils subclass

    model_type : string
        Must be one of [PERISOMATIC_TYPE, ALL_ACTIVE_TYPE].  If none, defaults to PERISOMATIC_TYPE

    Returns
    -------
    Utils instance    
    '''

    

    if model_type is None:
        try:
            model_type = description.data['biophys'][0]['model_type']
        except KeyError as e:
            logging.error("Could not infer model type from description")

    if model_type == PERISOMATIC_TYPE:
        return Utils(description)
    elif model_type == ALL_ACTIVE_TYPE:
        return AllActiveUtils(description)


class Utils(HocUtils):
    '''A helper class for NEURON functionality needed for
    biophysical simulations.

    Attributes
    ----------
    h : object
        The NEURON hoc object.
    nrn : object
        The NEURON python object.
    neuron : module
        The NEURON module.
    '''

    _log = logging.getLogger(__name__)

    def __init__(self, description):                    
        self.update_default_cell_hoc(description)

        super(Utils, self).__init__(description)
        self.stim = None
        self.stim_curr = None
        self.simulation_sampling_rate = None
        self.stimulus_sampling_rate = None

        self.stim_vec_list = []
        

    def update_default_cell_hoc(self, description, default_cell_hoc='cell.hoc'):
        ''' replace the default 'cell.hoc' path in the manifest with 'cell.hoc' packaged
        within AllenSDK if it does not exist '''

        hoc_files = description.data['neuron'][0]['hoc']
        try:
            hfi = hoc_files.index(default_cell_hoc)

            if not os.path.exists(default_cell_hoc):
                abspath_ch = resource_filename(__name__, 
                                               default_cell_hoc)
                hoc_files[hfi] = abspath_ch

                if not os.path.exists(abspath_ch):
                    raise IOError("cell.hoc does not exist!")

                self._log.warning("Using cell.hoc from the following location: %s", abspath_ch)
        except ValueError as e:
            pass


    def generate_morphology(self, morph_filename):
        '''Load a swc-format cell morphology file.

        Parameters
        ----------
        morph_filename : string
            Path to swc.
        '''
        h = self.h

        swc = self.h.Import3d_SWC_read()
        swc.input(morph_filename)
        imprt = self.h.Import3d_GUI(swc, 0)

        h("objref this")
        imprt.instantiate(h.this)

        h("soma[0] area(0.5)")
        for sec in h.allsec():
            sec.nseg = 1 + 2 * int(sec.L / 40.0)
            if sec.name()[:4] == "axon":
                h.delete_section(sec=sec)
        h('create axon[2]')
        for sec in h.axon:
            sec.L = 30
            sec.diam = 1
            sec.nseg = 1 + 2 * int(sec.L / 40.0)
        h.axon[0].connect(h.soma[0], 0.5, 0.0)
        h.axon[1].connect(h.axon[0], 1.0, 0.0)

        h.define_shape()

    def load_cell_parameters(self):
        '''Configure a neuron after the cell morphology has been loaded.'''
        passive = self.description.data['passive'][0]
        genome = self.description.data['genome']
        conditions = self.description.data['conditions'][0]
        h = self.h

        h("access soma")

        # Set fixed passive properties
        for sec in h.allsec():
            sec.Ra = passive['ra']
            sec.insert('pas')
            for seg in sec:
                seg.pas.e = passive["e_pas"]

        for c in passive["cm"]:
            h('forsec "' + c["section"] + '" { cm = %g }' % c["cm"])

        # Insert channels and set parameters
        for p in genome:
            if p["section"] == "glob":  # global parameter
                h(p["name"] + " = %g " % p["value"])
            else:
                if p["mechanism"] != "":
                    h('forsec "' + p["section"] +
                      '" { insert ' + p["mechanism"] + ' }')
                h('forsec "' + p["section"] +
                  '" { ' + p["name"] + ' = %g }' % p["value"])

        # Set reversal potentials
        for erev in conditions['erev']:
            h('forsec "' + erev["section"] + '" { ek = %g }' % erev["ek"])
            h('forsec "' + erev["section"] + '" { ena = %g }' % erev["ena"])

    def setup_iclamp(self,
                     stimulus_path,
                     sweep=0):
        '''Assign a current waveform as input stimulus. 

        Parameters
        ----------
        stimulus_path : string
            NWB file name
        '''
        self.stim = self.h.IClamp(self.h.soma[0](0.5))
        self.stim.amp = 0
        self.stim.delay = 0
        # just set to be really big; doesn't need to match the waveform
        self.stim.dur = 1e12

        self.read_stimulus(stimulus_path, sweep=sweep)

        # NEURON's dt is in milliseconds
        simulation_dt = 1.0e3 / self.simulation_sampling_rate
        stimulus_dt = 1.0e3 / self.stimulus_sampling_rate
        self._log.debug("Using simulation dt %f, stimulus dt %f", simulation_dt, stimulus_dt)

        self.h.dt = simulation_dt
        stim_vec = self.h.Vector(self.stim_curr)
        stim_vec.play(self.stim._ref_amp, stimulus_dt)

        stimulus_stop_index = len(self.stim_curr) - 1
        self.h.tstop = stimulus_stop_index * stimulus_dt
        self.stim_vec_list.append(stim_vec)

    def read_stimulus(self, stimulus_path, sweep=0):
        '''Load current values for a specific experiment sweep and setup simulation
        and stimulus sampling rates.

        NOTE: NEURON only allows simulation timestamps of multiples of 40KHz.  To 
        avoid aliasing, we set the simulation sampling rate to the least common
        multiple of the stimulus sampling rate and 40KHz.

        Parameters
        ----------
        stimulus path : string
            NWB file name
        sweep : integer, optional
            sweep index
        '''
        Utils._log.info(
            "reading stimulus path: %s, sweep %s",
            stimulus_path,
            sweep)

        stimulus_data = NwbDataSet(stimulus_path)
        sweep_data = stimulus_data.get_sweep(sweep)

        # convert to nA for NEURON
        self.stim_curr = sweep_data['stimulus'] * 1.0e9

        # convert from Hz
        hz = int(sweep_data['sampling_rate'])
        neuron_hz = Utils.nearest_neuron_sampling_rate(hz)

        self.simulation_sampling_rate = neuron_hz
        self.stimulus_sampling_rate = hz

        if hz != neuron_hz:
            Utils._log.debug("changing sampling rate from %d to %d to avoid NEURON aliasing", hz, neuron_hz)

    def record_values(self):
        '''Set up output voltage recording.'''
        vec = {"v": self.h.Vector(),
               "t": self.h.Vector()}

        vec["v"].record(self.h.soma[0](0.5)._ref_v)
        vec["t"].record(self.h._ref_t)

        return vec

    def get_recorded_data(self, vec):
        '''Extract recorded voltages and timestamps given the recorded Vector instance.  
        If self.stimulus_sampling_rate is smaller than self.simulation_sampling_rate, 
        resample to self.stimulus_sampling_rate.

        Parameters
        ----------
        vec : neuron.Vector 
           constructed by self.record_values

        Returns
        -------
        dict with two keys: 'v' = numpy.ndarray with voltages, 't' = numpy.ndarray with timestamps

        '''
        junction_potential = self.description.data['fitting'][0]['junction_potential']
        
        v = np.array(vec["v"])
        t = np.array(vec["t"])

        if self.stimulus_sampling_rate < self.simulation_sampling_rate:
            factor = self.simulation_sampling_rate / self.stimulus_sampling_rate
                
            Utils._log.debug("subsampling recorded traces by %dX", factor)
            v = block_reduce(v, (factor,), np.mean)[:len(self.stim_curr)]
            t = block_reduce(t, (factor,), np.min)[:len(self.stim_curr)]

        mV = 1.0e-3
        v = (v - junction_potential) * mV
        
        return { "v": v, "t": t }

    @staticmethod
    def nearest_neuron_sampling_rate(hz, target_hz=40000):
        div = gcd(hz, target_hz)
        new_hz = hz * target_hz / div
        return new_hz


class AllActiveUtils(Utils):

    def generate_morphology(self, morph_filename):
        '''Load a neurolucida or swc-format cell morphology file.

        Parameters
        ----------
        morph_filename : string
            Path to morphology.
        '''

        morph_basename = os.path.basename(morph_filename)
        morph_extension = morph_basename.split('.')[-1]
        if morph_extension.lower() == 'swc':
            morph = self.h.Import3d_SWC_read()
        elif morph_extension.lower() == 'asc':
            morph = self.h.Import3d_Neurolucida3()
        else:
            raise Exception("Unknown filetype: %s" % morph_extension)

        morph.input(morph_filename)
        imprt = self.h.Import3d_GUI(morph, 0)

        self.h("objref this")
        imprt.instantiate(self.h.this)

        for sec in self.h.allsec():
            sec.nseg = 1 + 2 * int(sec.L / 40.0)

        self.h("soma[0] area(0.5)")
        axon_diams = [self.h.axon[0].diam, self.h.axon[0].diam]
        self.h.distance(sec=self.h.soma[0])
        for sec in self.h.allsec():
            if sec.name()[:4] == "axon":
                if self.h.distance(0.5, sec=sec) > 60:
                    axon_diams[1] = sec.diam
                    break
        for sec in self.h.allsec():
            if sec.name()[:4] == "axon":
                self.h.delete_section(sec=sec)
        self.h('create axon[2]')
        for index, sec in enumerate(self.h.axon):
            sec.L = 30
            sec.diam = axon_diams[index]

        for sec in self.h.allsec():
            sec.nseg = 1 + 2 * int(sec.L / 40.0)

        self.h.axon[0].connect(self.h.soma[0], 1.0, 0.0)
        self.h.axon[1].connect(self.h.axon[0], 1.0, 0.0)

        # make sure diam reflects 3d points
        self.h.area(.5, sec=self.h.soma[0])

    def load_cell_parameters(self):
        '''Configure a neuron after the cell morphology has been loaded.'''
        passive = self.description.data['passive'][0]
        genome = self.description.data['genome']
        conditions = self.description.data['conditions'][0]
        h = self.h

        h("access soma")

        # Set fixed passive properties
        for sec in h.allsec():
            sec.Ra = passive['ra']
            sec.insert('pas')
            # for seg in sec:
            #     seg.pas.e = passive["e_pas"]

        # for c in passive["cm"]:
        #     h('forsec "' + c["section"] + '" { cm = %g }' % c["cm"])

        # Insert channels and set parameters
        for p in genome:
            section_array = p["section"]
            mechanism = p["mechanism"]
            param_name = p["name"]
            param_value = float(p["value"])
            if section_array == "glob":  # global parameter
                h(p["name"] + " = %g " % p["value"])
            else:
                if hasattr(h, section_array):
                    if mechanism != "":
                        print('Adding mechanism %s to %s'
                              % (mechanism, section_array))
                        for section in getattr(h, section_array):
                            if self.h.ismembrane(str(mechanism),
                                                 sec=section) != 1:
                                section.insert(mechanism)

                    print('Setting %s to %.6g in %s'
                          % (param_name, param_value, section_array))
                    for section in getattr(h, section_array):
                        setattr(section, param_name, param_value)

        # Set reversal potentials
        for erev in conditions['erev']:
            erev_section_array = erev["section"]
            ek = float(erev["ek"])
            ena = float(erev["ena"])

            print('Setting ek to %.6g and ena to %.6g in %s'
                  % (ek, ena, erev_section_array))

            if hasattr(h, erev_section_array):
                for section in getattr(h, erev_section_array):
                    if self.h.ismembrane("k_ion", sec=section) == 1:
                        setattr(section, 'ek', ek)

                    if self.h.ismembrane("na_ion", sec=section) == 1:
                        setattr(section, 'ena', ena)
            else:
                print("Warning: can't set erev for %s, "
                      "section array doesn't exist" % erev_section_array)

        self.h.v_init = conditions['v_init']
        self.h.celsius = conditions['celsius']



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import allensdk.brain_observatory.dff as dff
import numpy as np
import pytest
from functools import partial
from matplotlib.pyplot import Figure
from mock import patch, MagicMock


def test_movingmode_fast():
    # check basic behavior
    x = np.array([0, 10, 0, 0, 20, 0, 0, 0, 30])
    kernelsize = 4
    y = np.zeros(x.shape)

    dff.movingmode_fast(x, kernelsize, y)

    assert np.all(y == 0)

    # check window edges
    x = np.array([0, 0, 1, 1, 2, 2, 3, 3])
    kernelsize = 2
    y = np.zeros(x.shape)

    dff.movingmode_fast(x, kernelsize, y)

    assert np.all(x == y)

    # check > 16 bit
    x = np.array([4097, 4097, 4097, 4097])
    kernelsize = 2
    y = np.zeros(x.shape)

    dff.movingmode_fast(x, kernelsize, y)

    assert np.all(y == 4097)

    # check floats
    x = np.array([0, 0, 1, 1, 2, 2, 3, 3], dtype=np.float32)
    kernelsize = 2
    y = np.zeros(x.shape)

    dff.movingmode_fast(x, kernelsize, y)

    assert np.all(x == y)


def test_compute_dff():
    x = np.array([[1, 5, 0, 0, 1, 10, 0, 0, 30, 5]])

    with pytest.warns(FutureWarning):
        dff.compute_dff(x)


def test_compute_dff_windowed_mode():
    x = np.array([[1, 5, -2, 3, 1, 10, 1, -2, 30, 5]])

    with pytest.raises(ValueError):
        dff.compute_dff_windowed_mode(x, mode_kernelsize=0)
    with pytest.raises(ValueError):
        dff.compute_dff_windowed_mode(x, mean_kernelsize=0)

    y = dff.compute_dff_windowed_mode(x)

    assert(y.shape == x.shape)


def test_compute_dff_windowed_median():
    x = np.array([[1, 5, -2, 3, 1, 10, 1, -2, 30, 5]], dtype=float)

    with pytest.raises(ValueError):
        dff.compute_dff_windowed_median(x, median_kernel_long=2)
    with pytest.raises(ValueError):
        dff.compute_dff_windowed_median(x, median_kernel_short=-5)
    with pytest.raises(ValueError):
        dff.compute_dff_windowed_median(x, noise_kernel_length=50)
    with pytest.raises(ValueError):
        dff.compute_dff_windowed_median(x)

    x = np.sin(np.arange(0, 200)).reshape(1,200)

    y = dff.compute_dff_windowed_median(x, median_kernel_long=101,
                                        median_kernel_short=11,
                                        noise_kernel_length=5)

    assert(y.shape == x.shape)

    noise_stds = []
    small_frames = []
    y = dff.compute_dff_windowed_median(x, median_kernel_long=101,
                                        median_kernel_short=11,
                                        noise_stds=noise_stds,
                                        n_small_baseline_frames=small_frames,
                                        noise_kernel_length=5)

    assert len(noise_stds) == 1
    assert len(small_frames) == 1


def test_calculate_dff():
    x = np.array([[1, 5, -2, 3, 1, 10, 1, -2, 30, 5]], dtype=float)

    with patch("os.makedirs") as mock_makedirs:
        with patch.object(Figure, "savefig") as mock_save:
            with patch.object(dff, "compute_dff_windowed_median",
                       return_value=x) as mock_computation:
                dff.calculate_dff(x)
    assert mock_makedirs.call_count == 0
    assert mock_save.call_count == 0
    mock_computation.assert_called_once_with(x)

    with patch("os.makedirs") as mock_makedirs:
        with patch.object(Figure, "savefig") as mock_save:
            mock_computation = MagicMock(return_value=x)
            dff.calculate_dff(x, dff_computation_cb=mock_computation,
                              save_plot_dir="./test")
    mock_makedirs.assert_called_once_with("./test")
    mock_save.assert_called_once()
    mock_computation.assert_called_once_with(x)

    x = np.sin(np.arange(0, 200)).reshape(1,200)

    noise_stds = []
    small_frames = []
    computation_cb = partial(dff.compute_dff_windowed_median,
                             median_kernel_long=101,
                             median_kernel_short=11,
                             noise_stds=noise_stds,
                             n_small_baseline_frames=small_frames,
                             noise_kernel_length=5)
    dff.calculate_dff(x, dff_computation_cb=computation_cb)
    assert len(noise_stds) == 1
    assert len(small_frames) == 1



import os
import json

import numpy as np
import pytest
from mock import patch
from allensdk.api.queries.biophysical_api import BiophysicalApi


@pytest.fixture
def neuronal_model_response():
    dirname = os.path.dirname(__file__)
    path = os.path.join(dirname, 'response_test_data', '472451419_response.json')

    with open(path, 'r') as jf:
        data = json.load(jf)

    return data


@pytest.fixture
def biophys_api():
    endpoint = 'http://twarehouse-backup'
    return BiophysicalApi(endpoint)


@pytest.mark.parametrize('model_id', [3])
@pytest.mark.parametrize('fmt', [None, 'json', 'xml'])
def test_build_rma(model_id, fmt, biophys_api):
    if fmt is None:
        fmt_exp = 'json'
        obt = biophys_api.build_rma(model_id)
    else:
        fmt_exp = fmt
        obt = biophys_api.build_rma(model_id, fmt_exp)

    exp = 'http://twarehouse-backup/api/v2/data/query.{}?'\
          'q=model::NeuronalModel,'\
          'rma::criteria,[id$eq{}],'\
          'neuronal_model_template(well_known_files(well_known_file_type)),'\
          'specimen(ephys_result(well_known_files(well_known_file_type)),'\
          'neuron_reconstructions(well_known_files(well_known_file_type)),ephys_sweeps),'\
          'well_known_files(well_known_file_type),'\
          'rma::include,neuronal_model_template(well_known_files(well_known_file_type)),'\
          'specimen(ephys_result(well_known_files(well_known_file_type)),'\
          'neuron_reconstructions(well_known_files(well_known_file_type)),ephys_sweeps),'\
          'well_known_files(well_known_file_type)'
    exp = exp.format(fmt_exp, model_id)

    assert obt == exp


def test_is_well_known_file_type(biophys_api):
    wkf = {'well_known_file_type': {'name': 'fish'}}

    assert(biophys_api.is_well_known_file_type(wkf, 'fish'))
    assert(not biophys_api.is_well_known_file_type(wkf, 'fowl'))


@patch.object(BiophysicalApi, "json_msg_query")
def test_get_neuronal_models(mock_json_msg_query, biophys_api):
    mck = biophys_api.get_neuronal_models([386049446,469753383])

    mock_json_msg_query.assert_called_once_with(
        "http://twarehouse-backup/api/v2/data/query.json?"
        "q=model::NeuronalModel,rma::criteria,[neuronal_model_template_id$in491455321,32923071],"
        "[specimen_id$in386049446,469753383],rma::options[num_rows$eq'all'][count$eqfalse]")


def test_read_json(biophys_api, neuronal_model_response):
    
    obt = biophys_api.read_json(neuronal_model_response)

    assert(obt['stimulus']['491198851'] == "386049444.nwb")
    assert(obt['morphology']['491459173'] == "Nr5a1-Cre_Ai14-177334.05.01.01_491459171_m.swc")
    assert(obt['fit']['497235805'] == '386049446_fit.json')
    assert(obt['marker']['496607103'] == 'Nr5a1-Cre_Ai14-177334.05.01.01_491459171_marker_m.swc')
    assert(obt['modfiles']['395337293'] == os.path.join('modfiles', 'SK.mod'))
    assert(np.allclose(biophys_api.sweeps, [42]))



import pytest
import os
import numpy as np
from mock import patch, MagicMock
from allensdk.test_utilities import temp_dir


@pytest.fixture
def mock_request():
    return MagicMock()


@pytest.mark.parametrize("ismount,base_path",[
    (True, os.path.normpath(os.path.join('/', 'dev', 'shm'))),
    (False, os.path.dirname(temp_dir.__file__))
])
@patch("numpy.random.randint", side_effect=([1, 2, 3, 4, 5, 6],
                                            [1, 2, 3, 4, 5, 7]))
@patch("os.listdir", return_value=["allensdk_test_123456"])
@patch("os.makedirs")
def test_tmp_dir(os_makedirs, os_listdir, randint,
                 mock_request, ismount, base_path):
    with patch("os.path.exists", return_value=True):
        with patch("os.path.ismount", return_value=ismount):
            path = temp_dir.temp_dir(mock_request)
    mock_request.addfinalizer.assert_called_once()
    expected_path = os.path.join(base_path, "allensdk_test_123457")
    os_makedirs.assert_called_once_with(expected_path)
    assert path == expected_path
    with patch("shutil.rmtree") as mock_rmtree:
        with patch("os.path.exists", return_value=True):
            with pytest.warns(UserWarning):
                # run the finalizer
                mock_request.addfinalizer.call_args[0][0]()
        mock_rmtree.assert_called_once_with(expected_path)
        mock_rmtree.reset_mock()
        with patch("os.path.exists", return_value=False):
            mock_request.addfinalizer.call_args[0][0]()
        mock_rmtree.assert_called_once_with(expected_path)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import Mock, patch
from allensdk.api.cache import cacheable, Cache
from allensdk.config.manifest import Manifest
import allensdk.core.json_utilities as ju
import pandas.io.json as pj
import pandas as pd
from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi as MCA


try:
    import StringIO
except:
    import io as StringIO


@pytest.fixture
def mca():
    return MCA()


@pytest.fixture
def cache():
    return Cache()


@pytest.mark.parametrize("file_exists", (True, False))
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch.object(Manifest, 'safe_mkdir')
def test_file_download_lazy(nrrd_read, safe_mkdir, mca, cache, file_exists):
    with patch.object(mca, "retrieve_file_over_http") as mock_retrieve:
        @cacheable(strategy='lazy',
                reader=nrrd_read,
                pathfinder=Cache.pathfinder(file_name_position=3,
                                            secondary_file_name_position=1))
        def download_volumetric_data(data_path,
                                    file_name,
                                    voxel_resolution=None,
                                    save_file_path=None,
                                    release=None,
                                    coordinate_framework=None):
            url = mca.build_volumetric_data_download_url(data_path,
                                                        file_name,
                                                        voxel_resolution,
                                                        release,
                                                        coordinate_framework)

            mca.retrieve_file_over_http(url, save_file_path)

        with patch('os.path.exists',
                Mock(name="os.path.exists",
                        return_value=file_exists)) as mkdir:
            nrrd_read.reset_mock()
            download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                    'annotation_10.nrrd',
                                    MCA.VOXEL_RESOLUTION_10_MICRONS,
                                    'volumetric.nrrd',
                                    MCA.CCF_2016,
                                    strategy='lazy')

        if file_exists:
            assert not mock_retrieve.called, 'server call not needed when file exists'
        else:
            mock_retrieve.assert_called_once_with(
                'http://download.alleninstitute.org/informatics-archive/annotation/ccf_2016/mouse_ccf/average_template/annotation_10.nrrd',
                'volumetric.nrrd')
        assert not safe_mkdir.called, 'safe_mkdir should not have been called.'
        nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists", (True, False))
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch.object(Manifest, 'safe_mkdir')
def test_file_download_server(nrrd_read, safe_mkdir, mca, cache, file_exists):
    with patch.object(mca, "retrieve_file_over_http") as mock_retrieve:
        @cacheable(reader=nrrd_read,
                pathfinder=Cache.pathfinder(file_name_position=3,
                                            secondary_file_name_position=1))
        def download_volumetric_data(data_path,
                                    file_name,
                                    voxel_resolution=None,
                                    save_file_path=None,
                                    release=None,
                                    coordinate_framework=None):
            url = mca.build_volumetric_data_download_url(data_path,
                                                        file_name,
                                                        voxel_resolution,
                                                        release,
                                                        coordinate_framework)

            mca.retrieve_file_over_http(url, save_file_path)

        with patch('os.path.exists',
                Mock(name="os.path.exists",
                        return_value=file_exists)) as mkdir:
            nrrd_read.reset_mock()
            
            download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                    'annotation_10.nrrd',
                                    MCA.VOXEL_RESOLUTION_10_MICRONS,
                                    'volumetric.nrrd',
                                    MCA.CCF_2016,
                                    strategy='create')

        mock_retrieve.assert_called_once_with(
            'http://download.alleninstitute.org/informatics-archive/annotation/ccf_2016/mouse_ccf/average_template/annotation_10.nrrd',
            'volumetric.nrrd')
        assert not safe_mkdir.called, 'safe_mkdir should not have been called.'
        nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists", (True, False))
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch.object(Manifest, 'safe_mkdir')
def test_file_download_cached_file(nrrd_read, safe_mkdir, mca, cache, file_exists):
    with patch.object(mca, "retrieve_file_over_http") as mock_retrieve:
        @cacheable(reader=nrrd_read,
                pathfinder=Cache.pathfinder(file_name_position=3,
                                            secondary_file_name_position=1))
        def download_volumetric_data(data_path,
                                    file_name,
                                    voxel_resolution=None,
                                    save_file_path=None,
                                    release=None,
                                    coordinate_framework=None):
            url = mca.build_volumetric_data_download_url(data_path,
                                                        file_name,
                                                        voxel_resolution,
                                                        release,
                                                        coordinate_framework)

            mca.retrieve_file_over_http(url, save_file_path)

        with patch('os.path.exists',
                Mock(name="os.path.exists",
                        return_value=file_exists)) as mkdir:
            nrrd_read.reset_mock()

            download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                    'annotation_10.nrrd',
                                    MCA.VOXEL_RESOLUTION_10_MICRONS,
                                    'volumetric.nrrd',
                                    MCA.CCF_2016,
                                    strategy='file')

        assert not mock_retrieve.called, 'server should not have been called'
        assert not safe_mkdir.called, 'safe_mkdir should not have been called.'
        nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists", (True, False))
@patch("nrrd.read", return_value=('mock_annotation_data',
                                  'mock_annotation_image'))
@patch.object(Manifest, 'safe_mkdir')
def test_file_kwarg(nrrd_read, safe_mkdir, mca, cache, file_exists):
    with patch.object(mca, "retrieve_file_over_http") as mock_retrieve:
        @cacheable(reader=nrrd_read,
                pathfinder=Cache.pathfinder(file_name_position=3,
                                            secondary_file_name_position=1,
                                            path_keyword='save_file_path'))
        def download_volumetric_data(data_path,
                                    file_name,
                                    voxel_resolution=None,
                                    save_file_path=None,
                                    release=None,
                                    coordinate_framework=None):
            url = mca.build_volumetric_data_download_url(data_path,
                                                        file_name,
                                                        voxel_resolution,
                                                        release,
                                                        coordinate_framework)

            mca.retrieve_file_over_http(url, save_file_path)

        with patch('os.path.exists',
                Mock(name="os.path.exists",
                        return_value=file_exists)) as mkdir:
            nrrd_read.reset_mock()

            download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                    'annotation_10.nrrd',
                                    MCA.VOXEL_RESOLUTION_10_MICRONS,
                                    'volumetric.nrrd',
                                    MCA.CCF_2016,
                                    strategy='file',
                                    save_file_path='file.nrrd' )

        assert not mock_retrieve.called, 'server should not have been called'
        assert not safe_mkdir.called, 'safe_mkdir should not have been called.'
        nrrd_read.assert_called_once_with('file.nrrd')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import itertools as it

import pytest
import mock

from scipy.stats import multivariate_normal
from skimage.transform import rotate
import skimage
import numpy as np

import allensdk.brain_observatory.receptive_field_analysis.fitgaussian2D as gauss


@pytest.fixture(scope='function')
def gaussian_pdf():

    def gpdf(mean, cov, axes, scale=1):

        rv = multivariate_normal( mean, cov )
        mesh = np.meshgrid( *axes, indexing='ij' )
        pos = np.rollaxis( np.array(mesh), 0, len( axes ) + 1 )

        out = rv.pdf( pos )
        out = out / np.amax(out) * scale

        return out, mesh

    return gpdf


@pytest.fixture(scope='function')
def domain_axes():

    start = 0
    stop = 201
    step = 1
    naxes = 2

    axes = [ np.arange(start, stop, step) for ii in range(naxes) ]    
    return axes


@pytest.fixture(scope='function')
def simple_fill():

    def do_fill(domain_axes, fn):

        arr = np.zeros([ len(da) for da in domain_axes ])
        for pt in it.product(*domain_axes):
            arr[pt] = fn(*pt)

        return arr

    return do_fill


@pytest.mark.parametrize('mean,cov,scale', [ [ [ 100, 100 ], [ 25, 25 ], 1 ],
                                                 [ [ 100, 100 ], [ 10, 25 ], 1 ],
                                                 [ [ 100, 110 ], [ 25, 25 ], 1 ],
                                                 [ [ 100, 110 ], [ 10, 25 ], 1 ],
                                                 [ [ 110, 100 ], [ 10, 25 ], 1 ] ])
def test_gaussian2D_norot(mean, cov, scale, gaussian_pdf, domain_axes, simple_fill):

    full_cov = [ [ cov[0], 0 ], [ 0, cov[1] ] ]
    exp, mesh = gaussian_pdf( mean, full_cov, domain_axes, scale )
    
    obt_fn = gauss.gaussian2D( scale, mean[0], mean[1], np.sqrt(cov[0]), np.sqrt(cov[1]), 0 )
    obt = simple_fill( domain_axes, obt_fn )

    assert( np.allclose( obt, exp ) )


# only providing independent cov - using rotation after the fact
@pytest.mark.skipif(skimage.__version__ < '0.11.1', reason='cannot rotate about non-center point before .11.1')
@pytest.mark.parametrize('mean,cov,scale,rot', [ [ [ 100, 100 ], [ 25, 25 ], 1, 0 ],
                                                 [ [ 100, 100 ], [ 10, 25 ], 1, 0 ],
                                                 [ [ 100, 110 ], [ 25, 25 ], 1, 0 ],
                                                 [ [ 100, 110 ], [ 10, 25 ], 1, 0 ],
                                                 [ [ 110, 100 ], [ 10, 25 ], 1, 0 ],
                                                 [ [ 100, 100 ], [ 25, 25 ], 1, 90 ],
                                                 [ [ 100, 100 ], [ 25, 20 ], 1, 180 ],
                                                 [ [ 100, 100 ], [ 30, 25 ], 1, -90 ],
                                                 [ [ 100, 110 ], [ 20, 15 ], 1, -45 ],
                                                 [ [ 100, 110 ], [ 20, 15 ], 1, 30 ], 
                                                 [ [ 100, 100 ], [ 15, 20 ], 1, 10 ],
                                                 [ [ 100, 100 ], [ 10, 25 ], 10, 0 ] ])
def test_gaussian2D(mean, cov, scale, rot, gaussian_pdf, domain_axes, simple_fill):

    full_cov = [ [ cov[0], 0 ], [ 0, cov[1] ] ]
    exp, mesh = gaussian_pdf( mean, full_cov, domain_axes, scale )

    if rot != 0:
        exp = rotate( exp, -rot, False, center=mean[::-1] ) # negative rotation
    
    obt_fn = gauss.gaussian2D( scale, mean[0], mean[1], np.sqrt(cov[0]), np.sqrt(cov[1]), rot )
    obt = simple_fill( domain_axes, obt_fn )

    if rot == 0:
        assert( np.allclose( obt, exp ) )
    else:
        assert( np.linalg.norm( obt - exp ) / np.linalg.norm(exp) < 10 ** -2 )


@pytest.mark.parametrize('mean,cov,scale', [ [ [ 100, 100 ], [ [1, 0 ], [0, 1] ], 1 ],
                                             [ [ 100, 150 ], [ [1, 0 ], [0, 1] ], 1 ],
                                             [ [ 125, 125 ], [ [1, 0 ], [0, 1] ], 1 ],
                                             [ [ 110, 100 ], [ [1, 0 ], [0, 1] ], 1 ],
                                             [ [ 90, 100 ], [ [1, 0 ], [0, 1] ], 1 ],
                                             [ [ 100, 100 ], [ [1, 0 ], [0, 1] ], 2 ],
                                             [ [ 100, 100 ], [ [5, 0 ], [0, 1] ], 1 ] ])
def test_moments2(mean, cov, scale, gaussian_pdf, domain_axes):

    pdf, mesh = gaussian_pdf( mean, cov, domain_axes, scale )
    mom_exp = np.array([ scale, 
                         mean[0], mean[1], 
                         np.sqrt(cov[1][1]), np.sqrt(cov[0][0]) ])

    mom_obt = gauss.moments2( pdf )

    assert( np.allclose( mom_obt[:-1], mom_exp ) )
    assert( mom_obt[-1] is None ) # TODO: why?
    

# we probably want to test rotation here at some point, but there is no way that it could work now, given
# that moments2 assumes independence ...
@pytest.mark.parametrize('mean,cov,scale', [ [ [ 100, 100 ], [ 25, 25 ], 1 ],
                                             [ [ 100, 100 ], [ 10, 25 ], 1 ],
                                             [ [ 100, 110 ], [ 25, 25 ], 1 ],
                                             [ [ 100, 110 ], [ 10, 25 ], 1 ],
                                             [ [ 110, 100 ], [ 10, 25 ], 1 ] ])
def test_fitgaussian2D(mean, cov, scale, gaussian_pdf, domain_axes):

    full_cov = [ [ cov[0], 0 ], [ 0, cov[1] ] ]
    img, mesh = gaussian_pdf( mean, full_cov, domain_axes, scale )

    obt = gauss.fitgaussian2D( img )
    exp = [ scale, mean[0], mean[1], np.sqrt(cov[0]), np.sqrt(cov[1]), 0 ]

    assert( np.allclose( exp, obt, atol=10**-3 ) )


def test_fitgaussian2D_failure():

    data = np.eye(10)

    res = mock.MagicMock()
    res.success = False
    res.status = 3
    res.message = 'foo'

    with mock.patch('scipy.optimize.minimize', return_value=res) as p:
        with pytest.raises( gauss.GaussianFitError ):
            gauss.fitgaussian2D(data)



import functools

import h5py
import pytest
import numpy as np

import allensdk.core.h5_utilities as h5_utilities


@pytest.fixture
def mem_h5(request):
    my_file = h5py.File('my_file.h5', driver='core', backing_store=False)

    def fin():
        my_file.close()
    request.addfinalizer(fin)

    return my_file


@pytest.fixture
def simple_h5(mem_h5):
    mem_h5.create_group('a')
    mem_h5.create_group('a/b')
    mem_h5.create_group('a/b/c')
    mem_h5.create_group('d')
    mem_h5.create_group('a/e')

    return mem_h5


@pytest.fixture
def simple_h5_with_datsets(simple_h5):
    simple_h5.create_dataset(name='/a/b/c/fish', data=np.eye(10))
    simple_h5.create_dataset(name='a/fowl', data=np.eye(15))
    simple_h5.create_dataset(name='a/b/mammal', data=np.eye(20))

    return simple_h5


def test_decode_bytes():

    inp = np.array([b'a', b'b', b'c'])
    obt = h5_utilities.decode_bytes(inp)

    assert(np.array_equal( obt, ['a', 'b', 'c'] ))


def test_traverse_h5_file(simple_h5):

    names = []
    def cb(name, node):
        names.append(name)
    h5_utilities.traverse_h5_file(cb, simple_h5)

    assert( set(names) == set(['a', 'a/b', 'a/b/c', 'd', 'a/e']) )


def test_locate_h5_objects(simple_h5):

    matcher_cb = functools.partial(h5_utilities.h5_object_matcher_relname_in, ['c', 'e'])
    matches = h5_utilities.locate_h5_objects(matcher_cb, simple_h5)

    match_names = [ match.name for match in matches ]
    assert( set(match_names) == set(['/a/e', '/a/b/c']) )


def test_keyed_locate_h5_objects(simple_h5):

    matcher_cbs = {
        'e': functools.partial(h5_utilities.h5_object_matcher_relname_in, ['e']),
        'c': functools.partial(h5_utilities.h5_object_matcher_relname_in, ['c']),
    }

    matches = h5_utilities.keyed_locate_h5_objects(matcher_cbs, simple_h5)
    assert( matches['e'].name == '/a/e' )
    assert( matches['c'].name == '/a/b/c' )


def test_load_datasets_by_relnames(simple_h5_with_datsets):

    relnames = ['fish', 'fowl', 'mammal']
    obt = h5_utilities.load_datasets_by_relnames(relnames, simple_h5_with_datsets, simple_h5_with_datsets['a/b'])

    assert( len(obt) == 2 )
    assert(np.allclose( obt['fish'], np.eye(10) ))
    assert(np.allclose( obt['mammal'], np.eye(20) ))
